/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ComboBoxEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.MultiSelectionEditingSupport;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages {@link EStructuralFeature}s
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class EStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * Map: Instance key -> {@link EStructuralFeatureDescriptor}s of the
	 * {@link IAnnotatedSpecification} managed by this {@link IAnnotationValueProvider}.
	 */
	final private Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap;

	/**
	 * Flag if multiple {@link EStructuralFeature}s are managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	protected boolean handlesMultipleEstructuralFeatures;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EAttribute}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EAttribute eAttribute) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();
		structuralFeatureDescriptorMap.put(DEFAULT_KEY,
				new EStructuralFeatureDescriptor(eAttribute));

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a single {@link EReference}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EReference eReference, EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();
		structuralFeatureDescriptorMap.put(DEFAULT_KEY, new EStructuralFeatureDescriptor(
				eReference, eReferenceScope));

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented by a number of {@link EStructuralFeature}s contained in a single
	 * {@link EClass}. The mapping of of instance keys to the corresponding
	 * {@link EStructuralFeature}s is explicitly provided in a {@link Map}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap) {

		super(annotatedSpecificationEClass);

		this.structuralFeatureDescriptorMap = structuralFeatureDescriptorMap;

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented all {@link EStructuralFeature}s defined locally in a single
	 * {@link EClass} (i.e., not considering inherited attributes). The instance keys, and the
	 * mapping to the corresponding {@link EStructuralFeature}s are automatically derived from the
	 * underlying {@link EClass} (instance keys = field names of {@link EStructuralFeature}).
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass) {

		this(annotatedSpecificationEClass, false,
				EStructuralFeatureDescriptor.EReferenceScope.MODEL);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} that
	 * whose storage is implemented all {@link EStructuralFeature}s defined locally in a single
	 * {@link EClass} (i.e., not considering inherited attributes). The instance keys, and the
	 * mapping to the corresponding {@link EStructuralFeature}s are automatically derived from the
	 * underlying {@link EClass} (instance keys = field names of {@link EStructuralFeature}).
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			boolean inheritedFeatures, EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();

		EList<EStructuralFeature> featureList;
		if(inheritedFeatures) {
			featureList = annotatedSpecificationEClass.getEAllStructuralFeatures();
		} else {
			featureList = annotatedSpecificationEClass.getEStructuralFeatures();
		}

		if(featureList.size() == 1) {
			if(featureList.get(0) instanceof EAttribute) {
				structuralFeatureDescriptorMap.put(DEFAULT_KEY, new EStructuralFeatureDescriptor(
						(EAttribute)featureList.get(0)));
			} else if(featureList.get(0) instanceof EReference) {
				structuralFeatureDescriptorMap.put(DEFAULT_KEY, new EStructuralFeatureDescriptor(
						(EReference)featureList.get(0), eReferenceScope));
			}
		} else {
			for(EStructuralFeature structuralFeature : featureList) {
				if(structuralFeature instanceof EAttribute) {
					structuralFeatureDescriptorMap.put(structuralFeature.getName(),
							new EStructuralFeatureDescriptor((EAttribute)structuralFeature));
				} else if(structuralFeature instanceof EReference) {
					structuralFeatureDescriptorMap.put(structuralFeature.getName(),
							new EStructuralFeatureDescriptor((EReference)structuralFeature,
									eReferenceScope));
				}
			}
		}

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Returns the {@link EStructuralFeature} for a given {@code instanceKey}, or throws
	 * an exception, if the key is unknown.
	 * 
	 */
	private EStructuralFeature getEStructuralFeature(T specification, String instanceKey)
			throws Exception {
		EStructuralFeature structuralFeature =
				structuralFeatureDescriptorMap.get(instanceKey)
						.getEStructuralFeature(specification);
		if(structuralFeature == null) {
			throw new Exception("Instance key " + instanceKey +
					" is unknown for annotation provider " +
					annotatedSpecificationEClass.getInstanceClassName() + ".");
		}

		return structuralFeature;
	}

	/** Returns the {@link EFactory} for the given {@link EStructuralFeature}. */
	private EFactory getEFactory(EStructuralFeature structuralFeature) {
		return structuralFeature.getEType().getEPackage().getEFactoryInstance();
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	private <V> void
			setAnnotationValueFromString(String value, T specification, String instanceKey)
					throws Exception {

		EStructuralFeature structuralFeature = getEStructuralFeature(specification, instanceKey);

		if(structuralFeature instanceof EAttribute) {
			specification.eSet(
					structuralFeature,
					getEFactory(structuralFeature).createFromString(
							((EAttribute)structuralFeature).getEAttributeType(), value));
		} else {
			throw new Exception(
					"setAnnotationValueFromString() is only availabe for EAttributes. Use setAnnotationValue(U value, T specification) instead.");
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification, String instanceKey) {
		return true;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		if(value instanceof String) {
			// In case, the input choice is presented in dynamically managed combo box (i.e., where
			// new value choices can be entered by the user, the empty String indicates that the
			// current value should be deleted, and a neighboring value from the set should be used.
			// In case the last occurrence of a value choice has been removed, it will be no longer
			// be offered in the combo box.
			if(((String)value).isEmpty()) {
				String rval =
						getAnnotationValueAndUpdateInputChoice((String)value, specification,
								instanceKey);
				if(rval != null) {
					value = (U)rval;
				}
			}

			setAnnotationValueFromString((String)value, specification, instanceKey);
		} else {
			specification.eSet(getEStructuralFeature(specification, instanceKey), value);
		}
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		// See comment in setAnnotationValue(U value, T specification)
		if((value == null || value.isEmpty())) {
			String rval = getAnnotationValueAndUpdateInputChoice(value, specification, instanceKey);
			if(rval != null) {
				value = rval;
			}
		}

		EStructuralFeature structuralFeature = getEStructuralFeature(specification, instanceKey);

		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification, instanceKey);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented for the annotation type " +
							structuralFeature.getEType().getName() +
							"(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return (U)specification.eGet(getEStructuralFeature(specification, instanceKey));
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification) throws Exception {
		return getAnnotationValue(specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, T specification, String instanceKey)
			throws Exception {

		EStructuralFeature structuralFeature = getEStructuralFeature(specification, instanceKey);

		EClassifier eType = null;
		try {
			eType = structuralFeature.getEType();
		} catch(Exception e) {
			// Ignore
		}

		if(structuralFeature.getUpperBound() == 0) {
			throw new Exception(
					"EStructuralValueProvider: feature multiplicity == 0 is not supported.");
		}

		if(structuralFeature.getUpperBound() == 1) {

			if(((!(eType instanceof EEnum)) && (!(eType instanceof EClass))) ||
					((eType instanceof EClass) && (structuralFeatureDescriptorMap.get(instanceKey)
							.getEReferenceScope() == null))) {

				return super.createEditingSupport(viewer, clazz, specification, instanceKey);
			}
			return new ComboBoxEditingSupport(viewer, clazz, specification, instanceKey,
					structuralFeatureDescriptorMap.get(instanceKey));
		}

		if(eType instanceof EClass || eType instanceof EEnum) {
			return new MultiSelectionEditingSupport(viewer, clazz, specification, instanceKey,
					structuralFeatureDescriptorMap.get(instanceKey));
		}

		throw new Exception(
				"For feature multiplicity > 1, EStructuralValueProvider has been implemented for EReferences and EEnumns, only .");

	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (use to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 */
	protected Object createStructuralFeatureInstance(String instanceKey) throws Exception {
		EStructuralFeature structuralFeature = getEStructuralFeature(null, instanceKey);

		EFactory factory = getEFactory(structuralFeature);
		if(structuralFeature instanceof EAttribute) {

			EDataType attributeEDataType = ((EAttribute)structuralFeature).getEAttributeType();

			if(structuralFeature.getDefaultValueLiteral() != null) {
				// Use default value if it has been specified
				return factory.createFromString(attributeEDataType,
						structuralFeature.getDefaultValueLiteral());
			}

			// Return data types default value
			return attributeEDataType.getDefaultValue();
		} else if(structuralFeature instanceof EReference) {
			return null;
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented for the annotation type " +
							structuralFeature.getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {

		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance(instanceKey);
		if(structuralFeatureVal != null) {
			specification.eSet(getEStructuralFeature(specification, instanceKey),
					structuralFeatureVal);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		if(handlesMultipleEstructuralFeatures) {
			return structuralFeatureDescriptorMap.keySet();
		}
		return super.getInstanceKeys(specification);
	}
}
